/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package
 * @author
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "HandUnitGuiPlugin.h"
#include "HandUnitConfigDialog.h"
#include <RobotAPI/gui-plugins/HandUnitPlugin/ui_HandUnitConfigDialog.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/observers/variant/SingleTypeVariantList.h>

// Qt headers
#include <Qt>
#include <QtGlobal>
#include <QPushButton>
#include <QLabel>
#include <QLineEdit>
#include <QMessageBox>
#include <QTimer>
#include <QtWidgets/QSlider>
#include <QtWidgets/QGridLayout>
#include <QtWidgets/QLabel>
#include <QString>

#include <cmath>

namespace armarx
{
    HandUnitGuiPlugin::HandUnitGuiPlugin()
    {
        addWidget<HandUnitWidget>();
    }

    HandUnitWidget::HandUnitWidget() :
        leftHandName("NOT SET YET"),
        rightHandName("NOT SET YET"),
        leftHandUnitProxyName(""),
        rightHandUnitProxyName(""),
        setLeftHandJointAnglesFlag(false)
    {
        // init gui
        ui.setupUi(getWidget());

        leftHandJointAngleUpdateTask = new PeriodicTask<HandUnitWidget>(this, &HandUnitWidget::setLeftHandJointAngles, 50);
        rightHandJointAngleUpdateTask = new PeriodicTask<HandUnitWidget>(this, &HandUnitWidget::setRightHandJointAngles, 50);
        updateInfoTimer = new QTimer(this);
    }


    void HandUnitWidget::onInitComponent()
    {
        usingProxy(leftHandUnitProxyName);
        usingProxy(rightHandUnitProxyName);
        //usingTopic(handName + "State");
        //ARMARX_WARNING << "Listening on Topic: " << handName + "State";
    }

    void HandUnitWidget::onConnectComponent()
    {
        initGUIJointFrames();
        updateInfoTimer->start(50);

        leftHandJointAngleUpdateTask->start();

        leftHandUnitProxy = getProxy<HandUnitInterfacePrx>(leftHandUnitProxyName);
        leftHandName = leftHandUnitProxy->getHandName();

        // @@@ In simulation hand is called 'Hand L'/'Hand R'. On 3b Hand is called 'TCP R'
        if (leftHandName != "Hand L" && leftHandName != "TCP L")
        {
            //QMessageBox::warning(NULL, "Hand not supported", QString("Hand with name \"") + QString::fromStdString(handName) + " \" is not suppored.");
            ARMARX_WARNING << "Hand with name \"" << leftHandName << "\" is not supported.";
        }

        //ui.labelInfo->setText(QString::fromStdString(handUnitProxyName + " :: " + handName));

        SingleTypeVariantListPtr leftHandPreshapeStrings = SingleTypeVariantListPtr::dynamicCast(leftHandUnitProxy->getShapeNames());
        QStringList leftHandList;
        int leftHandPreshapeCount = leftHandPreshapeStrings->getSize();

        for (int i = 0; i < leftHandPreshapeCount; ++i)
        {
            std::string shape = ((leftHandPreshapeStrings->getVariant(i))->get<std::string>());
            leftHandList << QString::fromStdString(shape);
        }

        ui.comboLeftHandPreshapes->clear();
        ui.comboLeftHandPreshapes->addItems(leftHandList);



        rightHandJointAngleUpdateTask->start();

        rightHandUnitProxy = getProxy<HandUnitInterfacePrx>(rightHandUnitProxyName);
        rightHandName = rightHandUnitProxy->getHandName();

        // @@@ In simulation hand is called 'Hand L'/'Hand R'. On 3b Hand is called 'TCP R'
        if (rightHandName != "Hand R" && rightHandName != "TCP R")
        {
            ARMARX_WARNING << "Hand with name \"" << rightHandName << "\" is not supported.";
        }

        SingleTypeVariantListPtr rightHandPreshapeStrings = SingleTypeVariantListPtr::dynamicCast(rightHandUnitProxy->getShapeNames());
        QStringList rightHandList;
        int rightHandPreshapeCount = rightHandPreshapeStrings->getSize();

        for (int i = 0; i < rightHandPreshapeCount; ++i)
        {
            std::string shape = ((rightHandPreshapeStrings->getVariant(i))->get<std::string>());
            rightHandList << QString::fromStdString(shape);
        }

        ui.comboRightHandPreshapes->clear();
        ui.comboRightHandPreshapes->addItems(rightHandList);
    }

    void HandUnitWidget::onDisconnectComponent()
    {
        leftHandJointAngleUpdateTask->stop();
        rightHandJointAngleUpdateTask->stop();
        updateInfoTimer->stop();
    }

    void HandUnitWidget::onExitComponent()
    {
    }

    QPointer<QDialog> HandUnitWidget::getConfigDialog(QWidget* parent)
    {
        if (!dialog)
        {
            dialog = new HandUnitConfigDialog(parent);
            //dialog->ui->editHandUnitName->setText(QString::fromStdString(handUnitProxyName));
            //dialog->ui->editHandName->setText(QString::fromStdString(handName));
        }

        return qobject_cast<HandUnitConfigDialog*>(dialog);
    }

    void HandUnitWidget::configured()
    {
        leftHandUnitProxyName = dialog->proxyFinder->getSelectedProxyName().toStdString();
        rightHandUnitProxyName = dialog->proxyFinder->getSelectedProxyName().toStdString();
    }

    void HandUnitWidget::preshapeLeftHand()
    {
        setLeftHandPreshape(ui.comboLeftHandPreshapes->currentText().toUtf8().data());
    }

    void HandUnitWidget::preshapeRightHand()
    {
        setRightHandPreshape(ui.comboRightHandPreshapes->currentText().toUtf8().data());
    }

    void HandUnitWidget::setLeftHandJointAngles()
    {    
        if (!leftHandUnitProxy)
        {
            ARMARX_WARNING << "invalid proxy";
            return;
        }

        if (!setLeftHandJointAnglesFlag)
        {
            return;
        }

        setLeftHandJointAnglesFlag = false;

        NameValueMap leftHandJa;
        NameValueMap currentLeftHandJointValues = leftHandUnitProxy->getCurrentJointValues();

        if (leftHandName == "Hand L EEF")
        {
            for (const auto& pair : currentLeftHandJointValues)
            {
                leftHandJa[pair.first] = ui.frameLeftHand->findChild<QSlider *>(pair.first.c_str())->value();
            }
        }
        else if (leftHandName == "Hand L" || leftHandName == "TCP L")
        {
            for (const auto& pair : currentLeftHandJointValues)
            {
                leftHandJa[pair.first] = ui.frameLeftHand->findChild<QSlider *>(pair.first.c_str())->value() * M_PI / 2;
            }
        }
        else
        {
            ARMARX_WARNING << "Left hand with name \"" << leftHandName << "\" is not supported.";
        }

        leftHandUnitProxy->setJointAngles(leftHandJa);
    }

    void HandUnitWidget::requestSetLeftHandJointAngles()
    {
        setLeftHandJointAnglesFlag = true;
    }

    void HandUnitWidget::openLeftHand()
    {
        setLeftHandPreshape("Open");
    }

    void HandUnitWidget::closeLeftHand()
    {
        setLeftHandPreshape("Close");
    }

    void HandUnitWidget::relaxLeftHand()
    {
        setLeftHandPreshape("Relax");
    }

    void HandUnitWidget::updateInfoLabel()
    {
        ui.labelInfoLeftHand->setText(QString::fromStdString(leftHandUnitProxyName + " :: " + leftHandName + " State: " + leftHandUnitProxy->describeHandState()));
        ui.labelInfoRightHand->setText(QString::fromStdString(rightHandUnitProxyName + " :: " + rightHandName + " State: " + rightHandUnitProxy->describeHandState()));
    }

    void HandUnitWidget::setLeftHandPreshape(std::string preshape)
    {
        ARMARX_INFO << "Setting new left hand shape: " << preshape;
        leftHandUnitProxy->setShape(preshape);
    }

    void HandUnitWidget::loadSettings(QSettings* settings)
    {
        leftHandUnitProxyName = settings->value("leftHandUnitProxyName", QString::fromStdString(leftHandUnitProxyName)).toString().toStdString();
        leftHandName = settings->value("leftHandName", QString::fromStdString(leftHandName)).toString().toStdString();
        rightHandUnitProxyName = settings->value("rightHandUnitProxyName", QString::fromStdString(rightHandUnitProxyName)).toString().toStdString();
        rightHandName = settings->value("rightHandName", QString::fromStdString(rightHandName)).toString().toStdString();
    }

    void HandUnitWidget::saveSettings(QSettings* settings)
    {
        settings->setValue("leftHandUnitProxyName", QString::fromStdString(leftHandUnitProxyName));
        settings->setValue("leftHandName", QString::fromStdString(leftHandName));
        settings->setValue("rightHandUnitProxyName", QString::fromStdString(rightHandUnitProxyName));
        settings->setValue("rightHandName", QString::fromStdString(rightHandName));
    }


    void HandUnitWidget::initGUIJointFrames()
    {
        NameValueMap currentLeftHandJointValues = leftHandUnitProxy->getCurrentJointValues();
        QGridLayout* gridLayout = new QGridLayout(ui.frameLeftHand);
        int frameLeftHandRowIdx = 0;

        connect(ui.buttonPreshapeLeftHand, SIGNAL(clicked()), this, SLOT(preshapeLeftHand()), Qt::UniqueConnection);
        connect(ui.buttonOpenLeftHand, SIGNAL(clicked()), this, SLOT(openLeftHand()), Qt::UniqueConnection);
        connect(ui.buttonCloseLeftHand, SIGNAL(clicked()), this, SLOT(closeLeftHand()), Qt::UniqueConnection);
        connect(ui.buttonRelaxLeftHand, SIGNAL(clicked()), this, SLOT(relaxLeftHand()), Qt::UniqueConnection);
        connect(ui.buttonSetLeftHandJointAngles, SIGNAL(clicked()), this, SLOT(requestSetLeftHandJointAngles()), Qt::UniqueConnection);

        for (const auto& pair : currentLeftHandJointValues)
        {
            QLabel* label = new QLabel(ui.frameLeftHand);
            label->setText(QString::fromUtf8(pair.first.c_str()));
            gridLayout->addWidget(label, frameLeftHandRowIdx, 0, 1, 1);

            QSlider* horizontalSlider = new QSlider(ui.frameLeftHand);
            horizontalSlider->setObjectName("horizontalSlider" + QString::fromUtf8(pair.first.c_str()));
            horizontalSlider->setMaximum(1);
            horizontalSlider->setOrientation(Qt::Horizontal);
            connect(horizontalSlider, SIGNAL(sliderMoved(int)), this, SLOT(requestSetLeftHandJointAngles()), Qt::UniqueConnection);
            gridLayout->addWidget(horizontalSlider, frameLeftHandRowIdx, 1, 1, 2);

            frameLeftHandRowIdx++;
        }

        ui.frameLeftHand->setLayout(gridLayout);



        // TODO: right hand



        connect(updateInfoTimer, SIGNAL(timeout()), this, SLOT(updateInfoLabel()));
    }





    void armarx::HandUnitWidget::reportHandShaped(const std::string& handName, const std::string& handShapeName, const Ice::Current&)
    {
        ARMARX_IMPORTANT << handName << ": " << handShapeName;
    }

    void armarx::HandUnitWidget::reportNewHandShapeName(const std::string& handName, const std::string& handShapeName, const Ice::Current&)
    {
        ARMARX_IMPORTANT << handName << ": " << handShapeName;
    }

    void armarx::HandUnitWidget::reportJointAngles(const armarx::NameValueMap& actualJointAngles, const Ice::Current& c)
    {

    }

    void armarx::HandUnitWidget::reportJointPressures(const armarx::NameValueMap& actualJointPressures, const Ice::Current& c)
    {

    }
}
