/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD/STL

// Simox
#include <SimoxUtility/algorithm/string.h>

// Header
#include "Factory.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/type/variant/All.h>
#include "Data.h"

namespace armarx::aron::typereader::xml
{
    //std::map<std::string, typereader::GenerateObjectInfo> ReaderFactory::allGeneratedPublicObjects;
    //std::map<std::string, typereader::GenerateIntEnumInfo> ReaderFactory::allGeneratedPublicIntEnums;

    type::VariantPtr ReaderFactory::create(const RapidXmlReaderNode& node, const Path& path)
    {
        static const std::map<std::string, type::Descriptor> String2Descriptor =
        {
            {constantes::LIST_TAG, type::Descriptor::eList},
            {constantes::OBJECT_TAG, type::Descriptor::eObject},
            {constantes::TUPLE_TAG, type::Descriptor::eTuple},
            {constantes::PAIR_TAG, type::Descriptor::ePair},
            {constantes::DICT_TAG, type::Descriptor::eDict},
            {constantes::NDARRAY_TAG, type::Descriptor::eNDArray},
            {constantes::MATRIX_TAG, type::Descriptor::eMatrix},
            {constantes::QUATERNION_TAG, type::Descriptor::eQuaternion},
            {constantes::POINT_CLOUD_TAG, type::Descriptor::ePointCloud},
            {constantes::POSITION_TAG, type::Descriptor::ePosition},
            {constantes::ORIENTATION_TAG, type::Descriptor::eOrientation},
            {constantes::POSE_TAG, type::Descriptor::ePose},
            {constantes::IMAGE_TAG, type::Descriptor::eImage},
            {constantes::INT_ENUM_TAG, type::Descriptor::eIntEnum},
            {constantes::INT_TAG, type::Descriptor::eInt},
            {constantes::LONG_TAG, type::Descriptor::eLong},
            {constantes::FLOAT_TAG, type::Descriptor::eFloat},
            {constantes::DOUBLE_TAG, type::Descriptor::eDouble},
            {constantes::STRING_TAG, type::Descriptor::eString},
            {constantes::BOOL_TAG, type::Descriptor::eBool},
            {constantes::TIME_TAG, type::Descriptor::eTime},
            {constantes::ANY_OBJECT_TAG, type::Descriptor::eAnyObject}
        };

        const std::string tag = simox::alg::to_lower(node.name());
        auto it = String2Descriptor.find(tag);
        auto descriptor = (it == String2Descriptor.end() ? type::Descriptor::eUnknown : it->second);

        switch(descriptor)
        {
            case type::Descriptor::eList: return createList(node, path);
            case type::Descriptor::eDict: return createDict(node, path);
            case type::Descriptor::eObject: return createObject(node, path);
            case type::Descriptor::eTuple: return createTuple(node, path);
            case type::Descriptor::ePair: return createPair(node, path);
            case type::Descriptor::eNDArray: return createNDArray(node, path);
            case type::Descriptor::eMatrix: return createMatrix(node, path);
            case type::Descriptor::eQuaternion: return createQuaternion(node, path);
            case type::Descriptor::eImage: return createImage(node, path);
            case type::Descriptor::ePointCloud: return createPointCloud(node, path);
            case type::Descriptor::ePosition: return createPosition(node, path);
            case type::Descriptor::eOrientation: return createOrientation(node, path);
            case type::Descriptor::ePose: return createPose(node, path);
            case type::Descriptor::eIntEnum: return createIntEnum(node, path);
            case type::Descriptor::eInt: return createInt(node, path);
            case type::Descriptor::eLong: return createLong(node, path);
            case type::Descriptor::eFloat: return createFloat(node, path);
            case type::Descriptor::eDouble: return createDouble(node, path);
            case type::Descriptor::eString: return createString(node, path);
            case type::Descriptor::eBool: return createBool(node, path);
            case type::Descriptor::eTime: return createTime(node, path);
            case type::Descriptor::eAnyObject: return createAnyObject(node, path);
            default: return findExistingObject(node.name());
        }
    }

    type::VariantPtr ReaderFactory::findExistingObject(const std::string& name) const
    {
        const auto public_intenum_it = allGeneratedPublicIntEnums.find(name);
        if (public_intenum_it != allGeneratedPublicIntEnums.end())
        {
            // copy the navigator
            auto v = type::Variant::FromAronDTO(*public_intenum_it->second.correspondingType->toAronDTO());
            return v;
        }

        const auto public_obj_it = allGeneratedPublicObjects.find(name);
        if (public_obj_it != allGeneratedPublicObjects.end())
        {
            // copy the navigator and set instantiation template args
            auto v = type::Variant::FromAronDTO(*public_obj_it->second.correspondingType->toAronDTO());
            return v;
        }

        const auto public_unk_it = std::find(allPreviouslyKnownPublicTypes.begin(), allPreviouslyKnownPublicTypes.end(), name);
        if(public_unk_it != allPreviouslyKnownPublicTypes.end())
        {
            // create an empty navigator
            auto v = std::make_shared<aron::type::Object>(*public_unk_it);
            return v;
        }

        // only works during generation process (not for top-level)
        const auto public_priv_it = std::find(allPreviouslyKnownPrivateTypes.begin(), allPreviouslyKnownPrivateTypes.end(), name);
        if(public_priv_it != allPreviouslyKnownPrivateTypes.end())
        {
            // create an empty navigator
            auto v = std::make_shared<aron::type::Object>(*public_priv_it);
            return v;
        }

        throw error::ValueNotValidException(__PRETTY_FUNCTION__, "Cannot find a valid object.", name);
    }

    type::Maybe ReaderFactory::getMaybe(const RapidXmlReaderNode& n) const
    {
        if (util::AttributeIsTrue(n, constantes::OPTIONAL_NAME))
        {
            return type::Maybe::eOptional;
        }
        if (util::AttributeIsTrue(n, constantes::RAW_PTR_NAME))
        {
            return type::Maybe::eRawPointer;
        }
        if (util::AttributeIsTrue(n, constantes::SHARED_PTR_NAME))
        {
            return type::Maybe::eSharedPointer;
        }
        if (util::AttributeIsTrue(n, constantes::UNIQUE_PTR_NAME))
        {
            return type::Maybe::eUniquePointer;
        }
        return type::Maybe::eNone;
    }

    void ReaderFactory::checkObjectMemberName(const std::string& s) const
    {
        if (simox::alg::starts_with(s, "_") || simox::alg::ends_with(s, "_"))
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "You used an invalid membername - '_' as starting or ending char is not allowed.", s);
        }

        if (simox::alg::starts_with(s, "aron"))
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "You used an invalid membername - The prefix 'aron' is used for codegeneration.", s);
        }
    }

    type::VariantPtr ReaderFactory::createObject(const RapidXmlReaderNode& node, const Path& path)
    {
        if (path.hasElement())
        {
            throw error::AronException(__PRETTY_FUNCTION__, "Having an inner class is not supported anymore since Aron Version 'beta 0.2.3'. Please move the inner class definition to the <" + constantes::GENERATE_TYPES_TAG + ">-tag.");
        }
        // ensured we are toplevel!

        util::EnforceAttribute(node, constantes::NAME_ATTRIBUTE_NAME);
        const std::string extends = util::GetAttributeWithDefault(node, constantes::EXTENDS_ATTRIBUTE_NAME, "");
        const std::string name = util::GetAttribute(node, constantes::NAME_ATTRIBUTE_NAME);

        auto templates = simox::alg::split(util::GetAttributeWithDefault(node, constantes::TEMPLATE_ATTRIBUTE_NAME, ""), ",");

        auto newObjectInfo = typereader::GenerateObjectInfo();
        newObjectInfo.typeName = name;
        newObjectInfo.doc_brief = util::GetAttributeWithDefault(node, constantes::DOC_BRIEF_NAME, "");
        newObjectInfo.doc_author = util::GetAttributeWithDefault(node, constantes::DOC_AUTHOR_NAME, "");

        // reset private known types
        allPreviouslyKnownPrivateTypes = templates;

        std::map<std::string, type::VariantPtr> members;
        for (const RapidXmlReaderNode& objectChild : node.nodes())
        {
            util::EnforceTagName(objectChild, constantes::OBJECT_CHILD_TAG);
            util::EnforceChildSize(objectChild, 1);

            util::EnforceAttribute(objectChild, constantes::KEY_ATTRIBUTE_NAME);
            const std::string key = util::GetAttribute(objectChild, constantes::KEY_ATTRIBUTE_NAME);

            checkObjectMemberName(key);

            if (util::HasAttribute(objectChild, constantes::DOC_BRIEF_NAME))
            {
                newObjectInfo.doc_members.insert({key, util::GetAttribute(objectChild, constantes::DOC_BRIEF_NAME)});
            }

            std::vector<RapidXmlReaderNode> children = objectChild.nodes();

            auto maybe = getMaybe(children[0]);
            type::VariantPtr childNavigator = create(children[0], path.withElement(key));

            if (childNavigator->getDescriptor() == aron::type::Descriptor::eObject)
            {
                // check if template args are present
                std::vector<std::string> templates = simox::alg::split(util::GetAttributeWithDefault(children[0], constantes::TEMPLATE_ATTRIBUTE_NAME, ""), ",");
                auto obj = aron::type::Object::DynamicCastAndCheck(childNavigator);
                for (const auto& t : templates)
                {
                    obj->addTemplateInstantiation(t);
                }
            }

            childNavigator->setMaybe(maybe);
            members.insert({key, childNavigator});
        }

        // set the new object
        auto aronObjectType = std::make_shared<type::Object>(name, templates, std::vector<std::string>(), members, path);

        if (extends != "")
        {
            auto parentObj = type::Object::DynamicCastAndCheck(findExistingObject(extends));
            aronObjectType->setExtends(parentObj);
        }

        newObjectInfo.correspondingType = aronObjectType;
        allGeneratedPublicObjects.emplace(newObjectInfo.typeName, newObjectInfo);

        allPreviouslyKnownPrivateTypes.clear();
        return aronObjectType;
    }

    type::VariantPtr ReaderFactory::createList(const RapidXmlReaderNode &node, const Path &path)
    {
        util::EnforceChildSize(node, 1);

        std::vector<RapidXmlReaderNode> c = node.nodes();
        const RapidXmlReaderNode typeNode = c[0];
        type::VariantPtr type = create(typeNode, path.withAcceptedType());
        type->setMaybe(getMaybe(typeNode));

        auto o = std::make_shared<type::List>(type, path);
        return o;
    }

    type::VariantPtr ReaderFactory::createDict(const RapidXmlReaderNode &node, const Path &path)
    {
        util::EnforceChildSize(node, 1);

        std::vector<RapidXmlReaderNode> c = node.nodes();
        const RapidXmlReaderNode typeNode = c[0];
        type::VariantPtr type = create(typeNode, path.withAcceptedType());
        type->setMaybe(getMaybe(typeNode));

        auto o = std::make_shared<type::Dict>(type, path);
        return o;
    }

    type::VariantPtr ReaderFactory::createTuple(const RapidXmlReaderNode& node, const Path& path)
    {
        util::EnforceChildSizeGreater(node, 1);

        unsigned int i = 0;
        std::vector<RapidXmlReaderNode> c = node.nodes();
        std::vector<type::VariantPtr> elementTypes;
        for (const RapidXmlReaderNode& tupleTypeDeclarationNode : c)
        {
            util::EnforceChildSize(tupleTypeDeclarationNode, 1);

            std::vector<RapidXmlReaderNode> typeNodeChildren = tupleTypeDeclarationNode.nodes();
            const RapidXmlReaderNode typeNode = typeNodeChildren[0];

            type::VariantPtr type = create(typeNode, path.withAcceptedTypeIndex(i++));
            type->setMaybe(getMaybe(typeNode));

            elementTypes.push_back(type);
        }

        auto o = std::make_shared<type::Tuple>(elementTypes, path);
        return o;
    }

    type::VariantPtr ReaderFactory::createPair(const RapidXmlReaderNode& node, const Path& path)
    {
        util::EnforceChildSize(node, 2);

        std::vector<RapidXmlReaderNode> c = node.nodes();
        const RapidXmlReaderNode type1Node = c[0];

        type::VariantPtr type1 = create(type1Node, path.withAcceptedTypeIndex(0));
        type1->setMaybe(getMaybe(type1Node));

        const RapidXmlReaderNode type2Node = c[1];
        type::VariantPtr type2 = create(type2Node, path.withAcceptedTypeIndex(1));
        type2->setMaybe(getMaybe(type2Node));

        auto o = std::make_shared<type::Pair>(type1, type2, path);
        return o;
    }

    type::VariantPtr ReaderFactory::createNDArray(const RapidXmlReaderNode& node, const Path& path) const
    {
        static const std::map<std::string, type::ndarray::ElementType> String2NDArrayType =
        {
            {"int8", type::ndarray::ElementType::int8},
            {"int16", type::ndarray::ElementType::int16},
            {"int32", type::ndarray::ElementType::int32},
            {"uint8", type::ndarray::ElementType::uint8},
            {"uint16", type::ndarray::ElementType::uint16},
            {"uint32", type::ndarray::ElementType::uint32},
            {"float32", type::ndarray::ElementType::float32},
            {"float64", type::ndarray::ElementType::float64}
        };

        return nullptr;
    }

    type::VariantPtr ReaderFactory::createMatrix(const RapidXmlReaderNode &node, const Path &path) const
    {
        static const std::map<std::string, type::matrix::ElementType> String2MatrixType =
        {
            {"int16", type::matrix::ElementType::int16},
            {"int32", type::matrix::ElementType::int32},
            {"int64", type::matrix::ElementType::int64},
            {"float32", type::matrix::ElementType::float32},
            {"float64", type::matrix::ElementType::float64}
        };

        auto o = std::make_shared<type::Matrix>(path);
        util::EnforceChildSize(node, 0);
        util::EnforceAttribute(node, constantes::TYPE_ATTRIBUTE_NAME);

        const int rows = std::stoi(util::GetAttributeWithDefault(node, constantes::ROWS_ATTRIBUTE_NAME, "4"));
        const int cols = std::stoi(util::GetAttributeWithDefault(node, constantes::COLS_ATTRIBUTE_NAME, "4"));
        std::string type = util::GetAttributeWithDefault(node, constantes::TYPE_ATTRIBUTE_NAME, "???");

        o->setRows(rows);
        o->setCols(cols);
        try
        {
            o->setElementType(String2MatrixType.at(simox::alg::to_lower(type)));
        }
        catch (std::out_of_range& e)
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "The type tag is not valid", type);
        }
        return o;
    }

    type::VariantPtr ReaderFactory::createQuaternion(const RapidXmlReaderNode &node, const Path &path) const
    {
        static const std::map<std::string, type::quaternion::ElementType> String2QuaternionType =
        {
            {"float32", type::quaternion::ElementType::float32},
            {"float64", type::quaternion::ElementType::float64}
        };

        auto o = std::make_shared<type::Quaternion>(path);
        util::EnforceChildSize(node, 0);
        util::EnforceAttribute(node, constantes::TYPE_ATTRIBUTE_NAME);

        std::string type = util::GetAttributeWithDefault(node, constantes::TYPE_ATTRIBUTE_NAME, "???");

        try
        {
            o->setElementType(String2QuaternionType.at(simox::alg::to_lower(type)));
        }
        catch (std::out_of_range& e)
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "The type tag is not valid", type);
        }
        return o;
    }


    type::VariantPtr ReaderFactory::createImage(const RapidXmlReaderNode &node, const Path &path) const
    {
        static const std::map<std::string, type::image::PixelType> String2PixelType =
        {
            {"rgb24", type::image::PixelType::rgb24},
            {"depth32", type::image::PixelType::depth32}
        };

        auto o = std::make_shared<type::Image>(path);
        util::EnforceChildSize(node, 0);
        util::EnforceAttribute(node, constantes::TYPE_ATTRIBUTE_NAME);

        std::string type = util::GetAttributeWithDefault(node, constantes::TYPE_ATTRIBUTE_NAME, "???");

        try
        {
            o->setPixelType(String2PixelType.at(simox::alg::to_lower(type)));
        }
        catch (std::out_of_range& e)
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "The type tag is not valid", type);
        }
        return o;
    }

    type::VariantPtr ReaderFactory::createPointCloud(const RapidXmlReaderNode &node, const Path &path) const
    {
        static const std::map<std::string, type::pointcloud::VoxelType> String2VoxelType =
        {
            {"pointxyz", type::pointcloud::VoxelType::PointXYZ},
            {"pointxyzi", type::pointcloud::VoxelType::PointXYZI},
            {"pointxyzl", type::pointcloud::VoxelType::PointXYZL},
            {"pointxyzrgb", type::pointcloud::VoxelType::PointXYZRGB},
            {"pointxyzrgbl", type::pointcloud::VoxelType::PointXYZRGBL},
            {"pointxyzrgba", type::pointcloud::VoxelType::PointXYZRGBA},
            {"pointxyzhsv", type::pointcloud::VoxelType::PointXYZHSV}
        };

        auto o = std::make_shared<type::PointCloud>(path);
        util::EnforceChildSize(node, 0);
        util::EnforceAttribute(node, constantes::TYPE_ATTRIBUTE_NAME);

        std::string type = util::GetAttributeWithDefault(node, constantes::TYPE_ATTRIBUTE_NAME, "???");
        try
        {
            o->setVoxelType(String2VoxelType.at(simox::alg::to_lower(type)));
        }
        catch (std::out_of_range& e)
        {
            throw error::ValueNotValidException(__PRETTY_FUNCTION__, "The type tag is not valid", type);
        }
        return o;
    }

    type::VariantPtr ReaderFactory::createPosition(const RapidXmlReaderNode &node, const Path &path) const
    {
        auto o = std::make_shared<type::Position>(path);
        util::EnforceChildSize(node, 0);

        return o;
    }

    type::VariantPtr ReaderFactory::createOrientation(const RapidXmlReaderNode &node, const Path &path) const
    {
        auto o = std::make_shared<type::Orientation>(path);
        util::EnforceChildSize(node, 0);

        return o;
    }

    type::VariantPtr ReaderFactory::createPose(const RapidXmlReaderNode& node, const Path& path) const
    {
        auto o = std::make_shared<type::Pose>(path);
        util::EnforceChildSize(node, 0);

        return o;
    }

    type::VariantPtr ReaderFactory::createIntEnum(const RapidXmlReaderNode& node, const Path& path)
    {
        if (path.hasElement())
        {
            throw error::AronException(__PRETTY_FUNCTION__, "Having an inner int-enum is not supported anymore since Aron Version 'beta 0.2.3'. Please move the inner int-enum definition to the <" + constantes::GENERATE_TYPES_TAG + ">-tag.");
        }
        // ensured we are top-level!

        const std::string name = util::GetAttribute(node, constantes::NAME_ATTRIBUTE_NAME);

        auto newEnumInfo = typereader::GenerateIntEnumInfo();
        newEnumInfo.typeName = name;

        allPreviouslyKnownPrivateTypes.clear();

        std::map<std::string, int> acceptedValues;
        for (const RapidXmlReaderNode& valueChild : node.nodes())
        {
            util::EnforceTagName(valueChild, constantes::ENUM_VALUE_TAG);
            util::EnforceChildSize(valueChild, 0);

            const std::string key = util::GetAttribute(valueChild, constantes::KEY_ATTRIBUTE_NAME);

            if (util::HasAttribute(valueChild, constantes::DOC_BRIEF_NAME))
            {
                newEnumInfo.doc_values.insert({key, util::GetAttribute(valueChild, constantes::DOC_BRIEF_NAME)});
            }

            const std::string value = util::GetAttribute(valueChild, constantes::VALUE_ATTRIBUTE_NAME);

            acceptedValues.emplace(key, std::stoi(value));
        }
        auto o = std::make_shared<type::IntEnum>(name, acceptedValues, path);
        newEnumInfo.correspondingType = o;

        allGeneratedPublicIntEnums.emplace(newEnumInfo.typeName, newEnumInfo);

        allPreviouslyKnownPrivateTypes.clear();
        return o;
    }

    type::VariantPtr ReaderFactory::createInt(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::Int>(path);
    }

    type::VariantPtr ReaderFactory::createLong(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::Long>(path);
    }

    type::VariantPtr ReaderFactory::createFloat(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::Float>(path);
    }

    type::VariantPtr ReaderFactory::createDouble(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::Double>(path);
    }

    type::VariantPtr ReaderFactory::createString(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::String>(path);
    }

    type::VariantPtr ReaderFactory::createBool(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::Bool>(path);
    }

    type::VariantPtr ReaderFactory::createTime(const RapidXmlReaderNode &node, const Path &path) const
    {
        return std::make_shared<type::Time>(path);
    }

    type::VariantPtr ReaderFactory::createAnyObject(const RapidXmlReaderNode& node, const Path& path) const
    {
        return std::make_shared<type::AnyObject>(path);
    }
}
